#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright (c) 2016, Silvio Peroni <essepuntato@gmail.com>
#
# Permission to use, copy, modify, and/or distribute this software for any purpose
# with or without fee is hereby granted, provided that the above copyright notice
# and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT,
# OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
# DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
# ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
# SOFTWARE.
from __future__ import annotations

from typing import TYPE_CHECKING
from rdflib import Graph, Namespace, URIRef

from oc_ocdm.abstract_entity import AbstractEntity

if TYPE_CHECKING:
    from typing import ClassVar, Dict, Tuple, List, Optional
    from oc_ocdm.graph.graph_set import GraphSet


class GraphEntity(AbstractEntity):
    BIRO: ClassVar[Namespace] = Namespace("http://purl.org/spar/biro/")
    C4O: ClassVar[Namespace] = Namespace("http://purl.org/spar/c4o/")
    CO: ClassVar[Namespace] = Namespace("http://purl.org/co/")
    CITO: ClassVar[Namespace] = Namespace("http://purl.org/spar/cito/")
    DATACITE: ClassVar[Namespace] = Namespace("http://purl.org/spar/datacite/")
    DCTERMS: ClassVar[Namespace] = Namespace("http://purl.org/dc/terms/")
    DEO: ClassVar[Namespace] = Namespace("http://purl.org/spar/deo/")
    DOCO: ClassVar[Namespace] = Namespace("http://purl.org/spar/doco/")
    FABIO: ClassVar[Namespace] = Namespace("http://purl.org/spar/fabio/")
    FOAF: ClassVar[Namespace] = Namespace("http://xmlns.com/foaf/0.1/")
    FRBR: ClassVar[Namespace] = Namespace("http://purl.org/vocab/frbr/core#")
    LITERAL: ClassVar[Namespace] = Namespace("http://www.essepuntato.it/2010/06/literalreification/")
    OA: ClassVar[Namespace] = Namespace("http://www.w3.org/ns/oa#")
    OCO: ClassVar[Namespace] = Namespace("https://w3id.org/oc/ontology/")
    PRISM: ClassVar[Namespace] = Namespace("http://prismstandard.org/namespaces/basic/2.0/")
    PRO: ClassVar[Namespace] = Namespace("http://purl.org/spar/pro/")

    iri_has_subtitle: ClassVar[URIRef] = FABIO.hasSubtitle
    iri_has_publication_date: ClassVar[URIRef] = PRISM.publicationDate
    iri_bibliographic_reference: ClassVar[URIRef] = BIRO.BibliographicReference
    iri_references: ClassVar[URIRef] = BIRO.references
    iri_denotes: ClassVar[URIRef] = C4O.denotes
    iri_has_content: ClassVar[URIRef] = C4O.hasContent
    iri_intextref_pointer: ClassVar[URIRef] = C4O.InTextReferencePointer
    iri_is_context_of: ClassVar[URIRef] = C4O.isContextOf
    iri_singleloc_pointer_list: ClassVar[URIRef] = C4O.SingleLocationPointerList
    iri_has_element: ClassVar[URIRef] = CO.element
    iri_citation: ClassVar[URIRef] = CITO.Citation
    iri_cites: ClassVar[URIRef] = CITO.cites
    iri_citation_characterisation: ClassVar[URIRef] = CITO.hasCitationCharacterisation
    iri_has_citing_entity: ClassVar[URIRef] = CITO.hasCitingEntity
    iri_has_cited_entity: ClassVar[URIRef] = CITO.hasCitedEntity
    iri_oci: ClassVar[URIRef] = DATACITE.oci
    iri_doi: ClassVar[URIRef] = DATACITE.doi
    iri_pmid: ClassVar[URIRef] = DATACITE.pmid
    iri_pmcid: ClassVar[URIRef] = DATACITE.pmcid
    iri_orcid: ClassVar[URIRef] = DATACITE.orcid
    iri_xpath: ClassVar[URIRef] = DATACITE["local-resource-identifier-scheme"]
    iri_intrepid: ClassVar[URIRef] = DATACITE["intrepid"]
    iri_xmlid: ClassVar[URIRef] = DATACITE["local-resource-identifier-scheme"]
    iri_has_identifier: ClassVar[URIRef] = DATACITE.hasIdentifier
    iri_identifier: ClassVar[URIRef] = DATACITE.Identifier
    iri_isbn: ClassVar[URIRef] = DATACITE.isbn
    iri_issn: ClassVar[URIRef] = DATACITE.issn
    iri_url: ClassVar[URIRef] = DATACITE.url
    iri_uses_identifier_scheme: ClassVar[URIRef] = DATACITE.usesIdentifierScheme
    iri_title: ClassVar[URIRef] = DCTERMS["title"]
    iri_caption: ClassVar[URIRef] = DEO.Caption
    iri_discourse_element: ClassVar[URIRef] = DEO.DiscourseElement
    iri_footnote: ClassVar[URIRef] = DOCO.Footnote
    iri_paragraph: ClassVar[URIRef] = DOCO.Paragraph
    iri_part: ClassVar[URIRef] = DOCO.Part
    iri_section: ClassVar[URIRef] = DOCO.Section
    iri_section_title: ClassVar[URIRef] = DOCO.SectionTitle
    iri_sentence: ClassVar[URIRef] = DOCO.Sentence
    iri_table: ClassVar[URIRef] = DOCO.Table
    iri_text_chunk: ClassVar[URIRef] = DOCO.TextChunk
    iri_academic_proceedings: ClassVar[URIRef] = FABIO.AcademicProceedings
    iri_book: ClassVar[URIRef] = FABIO.Book
    iri_book_chapter: ClassVar[URIRef] = FABIO.BookChapter
    iri_book_series: ClassVar[URIRef] = FABIO.BookSeries
    iri_book_set: ClassVar[URIRef] = FABIO.BookSet
    iri_data_file: ClassVar[URIRef] = FABIO.DataFile
    iri_expression: ClassVar[URIRef] = FABIO.Expression
    iri_expression_collection: ClassVar[URIRef] = FABIO.ExpressionCollection
    iri_has_sequence_identifier: ClassVar[URIRef] = FABIO.hasSequenceIdentifier
    iri_journal: ClassVar[URIRef] = FABIO.Journal
    iri_journal_article: ClassVar[URIRef] = FABIO.JournalArticle
    iri_journal_issue: ClassVar[URIRef] = FABIO.JournalIssue
    iri_journal_volume: ClassVar[URIRef] = FABIO.JournalVolume
    iri_manifestation: ClassVar[URIRef] = FABIO.Manifestation
    iri_proceedings_paper: ClassVar[URIRef] = FABIO.ProceedingsPaper
    iri_reference_book: ClassVar[URIRef] = FABIO.ReferenceBook
    iri_reference_entry: ClassVar[URIRef] = FABIO.ReferenceEntry
    iri_report_document: ClassVar[URIRef] = FABIO.ReportDocument
    iri_series: ClassVar[URIRef] = FABIO.Series
    iri_specification_document: ClassVar[URIRef] = FABIO.SpecificationDocument
    iri_thesis: ClassVar[URIRef] = FABIO.Thesis
    iri_agent: ClassVar[URIRef] = FOAF.Agent
    iri_family_name: ClassVar[URIRef] = FOAF.familyName
    iri_given_name: ClassVar[URIRef] = FOAF.givenName
    iri_name: ClassVar[URIRef] = FOAF.name
    iri_embodiment: ClassVar[URIRef] = FRBR.embodiment
    iri_part_of: ClassVar[URIRef] = FRBR.partOf
    iri_contains_reference: ClassVar[URIRef] = FRBR.part
    iri_contains_de: ClassVar[URIRef] = FRBR.part
    iri_has_literal_value: ClassVar[URIRef] = LITERAL.hasLiteralValue
    iri_ending_page: ClassVar[URIRef] = PRISM.endingPage
    iri_starting_page: ClassVar[URIRef] = PRISM.startingPage
    iri_author: ClassVar[URIRef] = PRO.author
    iri_editor: ClassVar[URIRef] = PRO.editor
    iri_is_held_by: ClassVar[URIRef] = PRO.isHeldBy
    iri_publisher: ClassVar[URIRef] = PRO.publisher
    iri_is_document_context_for: ClassVar[URIRef] = PRO.isDocumentContextFor
    iri_role_in_time: ClassVar[URIRef] = PRO.RoleInTime
    iri_with_role: ClassVar[URIRef] = PRO.withRole
    iri_note: ClassVar[URIRef] = OA.Annotation
    iri_has_body: ClassVar[URIRef] = OA.hasBody
    iri_has_annotation: ClassVar[URIRef] = OCO.hasAnnotation  # inverse of OA.hasTarget
    iri_has_next: ClassVar[URIRef] = OCO.hasNext
    iri_archival_document: ClassVar[URIRef] = FABIO.ArchivalDocument
    iri_viaf: ClassVar[URIRef] = DATACITE.viaf
    iri_crossref: ClassVar[URIRef] = DATACITE.crossref  # TODO: add to datacite!
    iri_wikidata: ClassVar[URIRef] = DATACITE.wikidata  # TODO: add to datacite!
    iri_wikipedia: ClassVar[URIRef] = DATACITE.wikipedia  # TODO: add to datacite!
    iri_has_edition: ClassVar[URIRef] = PRISM.edition
    iri_relation: ClassVar[URIRef] = DCTERMS.relation
    iri_has_citation_creation_date: ClassVar[URIRef] = CITO.hasCitationCreationDate
    iri_has_citation_time_span: ClassVar[URIRef] = CITO.hasCitationTimeSpan
    iri_digital_manifestation: ClassVar[URIRef] = FABIO.DigitalManifestation
    iri_print_object: ClassVar[URIRef] = FABIO.PrintObject
    iri_has_url: ClassVar[URIRef] = FRBR.exemplar
    iri_self_citation: ClassVar[URIRef] = CITO.SelfCitation
    iri_affiliation_self_citation: ClassVar[URIRef] = CITO.AffiliationSelfCitation
    iri_author_network_self_citation: ClassVar[URIRef] = CITO.AuthorNetworkSelfCitation
    iri_author_self_citation: ClassVar[URIRef] = CITO.AuthorSelfCitation
    iri_funder_self_citation: ClassVar[URIRef] = CITO.FunderSelfCitation
    iri_journal_self_citation: ClassVar[URIRef] = CITO.JournalSelfCitation
    iri_journal_cartel_citation: ClassVar[URIRef] = CITO.JournalCartelCitation
    iri_distant_citation: ClassVar[URIRef] = CITO.DistantCitation
    iri_has_format: ClassVar[URIRef] = DCTERMS["format"]

    short_name_to_type_iri: ClassVar[Dict[str, URIRef]] = {
        'an': iri_note,
        'ar': iri_role_in_time,
        'be': iri_bibliographic_reference,
        'br': iri_expression,
        'ci': iri_citation,
        'de': iri_discourse_element,
        'id': iri_identifier,
        'pl': iri_singleloc_pointer_list,
        'ra': iri_agent,
        're': iri_manifestation,
        'rp': iri_intextref_pointer
    }

    def __init__(self, g: Graph, g_set: GraphSet, res: URIRef = None, res_type: URIRef = None,
                 resp_agent: str = None, source_agent: str = None, source: str = None,
                 count: str = None, label: str = None, short_name: str = "",
                 preexisting_graph: Graph = None) -> None:
        super(GraphEntity, self).__init__()
        self.g: Graph = g
        self.resp_agent: str = resp_agent
        self.source_agent: str = source_agent
        self.source: str = source
        self.short_name: str = short_name
        self.g_set: GraphSet = g_set
        self.preexisting_graph: Graph = Graph(identifier=g.identifier)
        self._merge_list: Tuple[GraphEntity] = tuple()
        # FLAGS
        self._to_be_deleted: bool = False
        self._was_merged: bool = False

        # If res was not specified, create from scratch the URI reference for this entity,
        # otherwise use the provided one
        if res is None:
            self.res = self._generate_new_res(g, count)
        else:
            self.res = res

        if g_set is not None:
            # If not already done, register this GraphEntity instance inside the GraphSet
            if self.res not in g_set.res_to_entity:
                g_set.res_to_entity[self.res] = self

        if preexisting_graph is not None:
            # Triples inside self.g are entirely replaced by triples from preexisting_graph.
            # This has maximum priority with respect to every other self.g initializations.
            # It's fundamental that the preexisting graph gets passed as an argument of the constructor:
            # allowing the user to set this value later through a method would mean that the user could
            # set the preexisting graph AFTER having modified self.g (which would not make sense).
            self.g.remove((None, None, None))
            for p, o in preexisting_graph.predicate_objects(self.res):
                self.g.add((self.res, p, o))
                self.preexisting_graph.add((self.res, p, o))
        else:
            # Add mandatory information to the entity graph
            self._create_type(res_type)
            if label is not None:
                self.create_label(label)

    @staticmethod
    def _generate_new_res(g: Graph, count: str) -> URIRef:
        return URIRef(str(g.identifier) + count)

    @property
    def to_be_deleted(self) -> bool:
        return self._to_be_deleted

    @property
    def was_merged(self) -> bool:
        return self._was_merged

    @property
    def merge_list(self) -> Tuple[GraphEntity]:
        return self._merge_list

    def mark_as_to_be_deleted(self, to_be_deleted: bool = True) -> None:
        self._to_be_deleted = to_be_deleted

    def merge(self, other: GraphEntity) -> None:
        types: List[URIRef] = other.get_types()
        for cur_type in types:
            self._create_type(cur_type)

        label: Optional[str] = other.get_label()
        if label is not None:
            self.create_label(label)

        self._was_merged = True
        other.mark_as_to_be_deleted()
        self._merge_list = (*self._merge_list, other)

    def commit_changes(self):
        self.preexisting_graph = Graph(identifier=self.g.identifier)
        if self._to_be_deleted:
            self.remove_every_triple()
        else:
            for triple in self.g.triples((None, None, None)):
                self.preexisting_graph.add(triple)
        self._to_be_deleted = False
        self._was_merged = False
        self._merge_list = tuple()



#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright (c) 2016, Silvio Peroni <essepuntato@gmail.com>
#
# Permission to use, copy, modify, and/or distribute this software for any purpose
# with or without fee is hereby granted, provided that the above copyright notice
# and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT,
# OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
# DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
# ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
# SOFTWARE.
from __future__ import annotations

from typing import TYPE_CHECKING

from oc_ocdm.decorators import accepts_only

if TYPE_CHECKING:
    from typing import Optional, List
    from rdflib import URIRef
    from oc_ocdm.graph.entities.bibliographic.reference_annotation import ReferenceAnnotation
    from oc_ocdm.graph.entities.bibliographic.bibliographic_resource import BibliographicResource
from oc_ocdm.graph.graph_entity import GraphEntity
from oc_ocdm.graph.entities.bibliographic_entity import BibliographicEntity


class BibliographicReference(BibliographicEntity):
    """Bibliographic reference (short: be): the particular textual bibliographic reference,
    usually occurring in the reference list (and denoted by one or more in-text reference
    pointers within the text of a citing bibliographic resource), that references another
    bibliographic resource.
    """

    @accepts_only('be')
    def merge(self, other: BibliographicReference) -> None:
        super(BibliographicReference, self).merge(other)

        content: Optional[str] = other.get_content()
        if content is not None:
            self.has_content(content)

        annotations_list: List[ReferenceAnnotation] = other.get_annotations()
        for cur_annotation in annotations_list:
            self.has_annotation(cur_annotation)

        referenced_br: Optional[BibliographicResource] = other.get_referenced_br()
        if referenced_br is not None:
            self.references_br(referenced_br)

    # HAS BIBLIOGRAPHIC REFERENCE TEXT
    def get_content(self) -> Optional[str]:
        return self._get_literal(GraphEntity.iri_has_content)

    @accepts_only('literal')
    def has_content(self, string: str) -> None:
        """The literal text of a bibliographic reference occurring in the reference list (or
        elsewhere) within a bibliographic resource, that references another bibliographic
        resource. The reference text should be recorded “as given” in the citing bibliographic
        resource, including any errors (e.g. mis-spellings of authors’ names, or changes from
        “β” in the original published title to “beta” in the reference text) or omissions (e.g.
        omission of the title of the referenced bibliographic resource, or omission of sixth and
        subsequent authors’ names, as required by certain publishers), and in whatever format
        it has been made available. For instance, the reference text can be either as plain text
        or as a block of XML.
        """
        self.remove_content()
        self._create_literal(GraphEntity.iri_has_content, string)

    def remove_content(self) -> None:
        self.g.remove((self.res, GraphEntity.iri_has_content, None))

    # HAS ANNOTATION (ReferenceAnnotation)
    def get_annotations(self) -> List[ReferenceAnnotation]:
        uri_list: List[URIRef] = self._get_multiple_uri_references(GraphEntity.iri_has_annotation)
        result: List[ReferenceAnnotation] = []
        for uri in uri_list:
            result.append(self.g_set.add_an(self.resp_agent, self.source_agent, self.source, uri))
        return result

    @accepts_only('an')
    def has_annotation(self, an_res: ReferenceAnnotation) -> None:
        """An annotation characterizing the related citation, in terms of its citation function (the
        reason for that citation).
        """
        self.g.add((self.res, GraphEntity.iri_has_annotation, an_res.res))

    @accepts_only('an')
    def remove_annotation(self, an_res: ReferenceAnnotation = None) -> None:
        if an_res is not None:
            self.g.remove((self.res, GraphEntity.iri_has_annotation, an_res.res))
        else:
            self.g.remove((self.res, GraphEntity.iri_has_annotation, None))

    # REFERENCES (BibliographicResource)
    def get_referenced_br(self) -> Optional[BibliographicResource]:
        uri: Optional[URIRef] = self._get_uri_reference(GraphEntity.iri_references)
        if uri is not None:
            return self.g_set.add_br(self.resp_agent, self.source_agent, self.source, uri)

    @accepts_only('br')
    def references_br(self, br_res: BibliographicResource) -> None:
        """The bibliographic reference that cites this bibliographic resource.
        """
        self.remove_referenced_br()
        self.g.add((self.res, GraphEntity.iri_references, br_res.res))

    def remove_referenced_br(self) -> None:
        self.g.remove((self.res, GraphEntity.iri_references, None))



#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright (c) 2016, Silvio Peroni <essepuntato@gmail.com>
#
# Permission to use, copy, modify, and/or distribute this software for any purpose
# with or without fee is hereby granted, provided that the above copyright notice
# and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT,
# OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
# DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
# ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
# SOFTWARE.
from __future__ import annotations

from typing import TYPE_CHECKING

from oc_ocdm.decorators import accepts_only

if TYPE_CHECKING:
    from typing import Optional, List
    from rdflib import URIRef
from oc_ocdm.graph.graph_entity import GraphEntity
from oc_ocdm.graph.entities.bibliographic_entity import BibliographicEntity


class ResponsibleAgent(BibliographicEntity):
    """Responsible agent (short: ra): the agent (usually a person or an organisation) having
       a certain role with respect to a bibliographic resource (e.g. an author of a paper or
       book, or an editor of a journal)."""

    @accepts_only('ra')
    def merge(self, other: ResponsibleAgent) -> None:
        super(ResponsibleAgent, self).merge(other)

        name: Optional[str] = other.get_name()
        if name is not None:
            self.has_name(name)

        given_name: Optional[str] = other.get_given_name()
        if given_name is not None:
            self.has_given_name(given_name)

        family_name: Optional[str] = other.get_family_name()
        if family_name is not None:
            self.has_family_name(family_name)

        related_agents: List[URIRef] = other.get_related_agents()
        for cur_agent in related_agents:
            self.has_related_agent(cur_agent)

    # HAS NAME STRING
    def get_name(self) -> Optional[str]:
        return self._get_literal(GraphEntity.iri_name)

    @accepts_only('literal')
    def has_name(self, string: str) -> None:
        """The name of an agent (for people, usually in the format: given name followed by family
        name, separated by a space).
        """
        self.remove_name()
        self._create_literal(GraphEntity.iri_name, string)

    def remove_name(self) -> None:
        self.g.remove((self.g, GraphEntity.iri_name, None))

    # HAS GIVEN NAME
    def get_given_name(self) -> Optional[str]:
        return self._get_literal(GraphEntity.iri_given_name)

    @accepts_only('literal')
    def has_given_name(self, string: str) -> None:
        """The given name of an agent, if a person.
        """
        self.remove_given_name()
        self._create_literal(GraphEntity.iri_given_name, string)

    def remove_given_name(self) -> None:
        self.g.remove((self.g, GraphEntity.iri_given_name, None))

    # HAS FAMILY NAME
    def get_family_name(self) -> Optional[str]:
        return self._get_literal(GraphEntity.iri_family_name)

    @accepts_only('literal')
    def has_family_name(self, string: str) -> None:
        """The family name of an agent, if a person.
        """
        self.remove_family_name()
        self._create_literal(GraphEntity.iri_family_name, string)

    def remove_family_name(self) -> None:
        self.g.remove((self.g, GraphEntity.iri_family_name, None))

    # HAS RELATED AGENT
    def get_related_agents(self) -> List[URIRef]:
        uri_list: List[URIRef] = self._get_multiple_uri_references(GraphEntity.iri_relation)
        return uri_list

    @accepts_only('thing')
    def has_related_agent(self, thing_ref: URIRef) -> None:
        """An external agent that/who is related in some relevant way with this responsible agent
        (e.g. for inter-linking purposes).
        """
        self.g.add((self.res, GraphEntity.iri_relation, thing_ref))

    @accepts_only('thing')
    def remove_related_agent(self, thing_ref: URIRef = None) -> None:
        if thing_ref is not None:
            self.g.remove((self.res, GraphEntity.iri_relation, thing_ref))
        else:
            self.g.remove((self.res, GraphEntity.iri_relation, None))



#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright (c) 2016, Silvio Peroni <essepuntato@gmail.com>
#
# Permission to use, copy, modify, and/or distribute this software for any purpose
# with or without fee is hereby granted, provided that the above copyright notice
# and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT,
# OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
# DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
# ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
# SOFTWARE.
from __future__ import annotations

from typing import TYPE_CHECKING

from oc_ocdm.decorators import accepts_only
from oc_ocdm.graph.graph_entity import GraphEntity
from oc_ocdm.prov.prov_entity import ProvEntity
from rdflib import XSD

if TYPE_CHECKING:
    from typing import Optional, List
    from rdflib import URIRef


class EntitySnapshot(ProvEntity):
    # HAS CREATION DATE
    def get_generation_time(self) -> Optional[str]:
        return self._get_literal(ProvEntity.iri_generated_at_time)

    @accepts_only('literal')
    def has_generation_time(self, string: str) -> None:
        """The date on which a particular snapshot of a bibliographic entity's metadata was
        created.
        """
        self.remove_generation_time()
        self._create_literal(ProvEntity.iri_generated_at_time, string, XSD.dateTime)

    def remove_generation_time(self) -> None:
        self.g.remove((self.res, ProvEntity.iri_generated_at_time, None))

    # HAS INVALIDATION DATE
    def get_invalidation_time(self) -> Optional[str]:
        return self._get_literal(ProvEntity.iri_invalidated_at_time)

    @accepts_only('literal')
    def has_invalidation_time(self, string: str) -> None:
        """The date on which a snapshot of a bibliographic entity's metadata was invalidated due
        to an update (e.g. a correction, or the addition of some metadata that was not specified
        in the previous snapshot), or due to a merger of the entity with another one.
        """
        self.remove_invalidation_time()
        self._create_literal(ProvEntity.iri_invalidated_at_time, string, XSD.dateTime)

    def remove_invalidation_time(self) -> None:
        self.g.remove((self.res, ProvEntity.iri_invalidated_at_time, None))

    # IS SNAPSHOT OF
    def get_is_snapshot_of(self) -> Optional[URIRef]:
        uri: Optional[URIRef] = self._get_uri_reference(ProvEntity.iri_specialization_of)
        return uri

    def is_snapshot_of(self, en_res: GraphEntity) -> None:
        """This property is used to link a snapshot of entity metadata to the bibliographic entity
        to which the snapshot refers.
        """
        self.remove_is_snapshot_of()
        self.g.add((self.res, ProvEntity.iri_specialization_of, en_res.res))

    def remove_is_snapshot_of(self) -> None:
        self.g.remove((self.res, ProvEntity.iri_specialization_of, None))

    # IS DERIVED FROM
    def get_derives_from(self) -> List[ProvEntity]:
        uri_list: List[URIRef] = self._get_multiple_uri_references(ProvEntity.iri_was_derived_from)
        result: List[ProvEntity] = []
        for uri in uri_list:
            # TODO: what is the prov_subject of these snapshots?
            result.append(self.p_set.add_se(None, uri))
        return result

    @accepts_only('se')
    def derives_from(self, se_res: ProvEntity) -> None:
        """This property is used to identify the immediately previous snapshot of entity metadata
        associated with the same bibliographic entity.
        """
        self.g.add((self.res, ProvEntity.iri_was_derived_from, se_res.res))

    @accepts_only('se')
    def remove_derives_from(self, se_res: ProvEntity = None) -> None:
        if se_res is not None:
            self.g.remove((self.res, ProvEntity.iri_was_derived_from, se_res.res))
        else:
            self.g.remove((self.res, ProvEntity.iri_was_derived_from, None))

    # HAS PRIMARY SOURCE
    def get_primary_source(self) -> Optional[URIRef]:
        uri: Optional[URIRef] = self._get_uri_reference(ProvEntity.iri_had_primary_source)
        return uri

    @accepts_only('thing')
    def has_primary_source(self, any_res: URIRef) -> None:
        """This property is used to identify the primary source from which the metadata
        described in the snapshot are derived (e.g. Crossref, as the result of querying the
        CrossRef API).
        """
        self.remove_primary_source()
        self.g.add((self.res, ProvEntity.iri_had_primary_source, any_res))

    def remove_primary_source(self) -> None:
        self.g.remove((self.res, ProvEntity.iri_had_primary_source, None))

    # HAS UPDATE ACTION
    def get_update_action(self) -> Optional[str]:
        return self._get_literal(ProvEntity.iri_has_update_query)

    @accepts_only('literal')
    def has_update_action(self, string: str) -> None:
        """The UPDATE SPARQL query that specifies which data, associated to the bibliographic
        entity in consideration, have been modified (e.g. for correcting a mistake) in the
        current snapshot starting from those associated to the previous snapshot of the entity.
        """
        self.remove_update_action()
        self._create_literal(ProvEntity.iri_has_update_query, string)

    def remove_update_action(self) -> None:
        self.g.remove((self.res, ProvEntity.iri_has_update_query, None))

    # HAS DESCRIPTION
    def get_description(self) -> Optional[str]:
        return self._get_literal(ProvEntity.iri_description)

    @accepts_only('literal')
    def has_description(self, string: str) -> None:
        """A textual description of the events that have resulted in the current snapshot (e.g. the
        creation of the initial snapshot, the creation of a new snapshot following the
        modification of the entity to which the metadata relate, or the creation of a new
        snapshot following the merger with another entity of the entity to which the previous
        snapshot related).
        """
        self.remove_description()
        self._create_literal(ProvEntity.iri_description, string)

    def remove_description(self) -> None:
        self.g.remove((self.res, ProvEntity.iri_description, None))

    # IS ATTRIBUTED TO
    def get_resp_agent(self) -> Optional[URIRef]:
        uri: Optional[URIRef] = self._get_uri_reference(ProvEntity.iri_was_attributed_to)
        return uri

    @accepts_only('thing')
    def has_resp_agent(self, se_agent: URIRef) -> None:
        """The agent responsible for the creation of the current entity snapshot.
        """
        self.remove_resp_agent()
        self.g.add((self.res, ProvEntity.iri_was_attributed_to, se_agent))

    def remove_resp_agent(self) -> None:
        self.g.remove((self.res, ProvEntity.iri_was_attributed_to, None))



#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright (c) 2016, Silvio Peroni <essepuntato@gmail.com>
#
# Permission to use, copy, modify, and/or distribute this software for any purpose
# with or without fee is hereby granted, provided that the above copyright notice
# and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT,
# OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
# DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
# ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
# SOFTWARE.
from __future__ import annotations

from typing import TYPE_CHECKING

from oc_ocdm.decorators import accepts_only
from oc_ocdm.metadata.metadata_entity import MetadataEntity
from rdflib import XSD

if TYPE_CHECKING:
    from typing import Optional, List
    from rdflib import URIRef
    from oc_ocdm.metadata.entities.distribution import Distribution


class Dataset(MetadataEntity):
    """Dataset (short: not applicable and strictly dependent on the implementation of the
       dataset infrastructure): a set of collected information about something."""

    @accepts_only('_dataset_')
    def merge(self, other: Dataset) -> None:
        super(Dataset, self).merge(other)

        title: Optional[str] = other.get_title()
        if title is not None:
            self.has_title(title)

        description: Optional[str] = other.get_description()
        if description is not None:
            self.has_description(description)

        pub_date: Optional[str] = other.get_publication_date()
        if pub_date is not None:
            self.has_publication_date(pub_date)

        mod_date: Optional[str] = other.get_modification_date()
        if mod_date is not None:
            self.has_modification_date(mod_date)

        keywords_list: List[str] = other.get_keywords()
        for cur_keyword in keywords_list:
            self.has_keyword(cur_keyword)

        subjects_list: List[URIRef] = other.get_subjects()
        for cur_subject in subjects_list:
            self.has_subject(cur_subject)

        landing_page: Optional[URIRef] = other.get_landing_page()
        if landing_page is not None:
            self.has_landing_page(landing_page)

        sub_datasets_list: List[Dataset] = other.get_sub_datasets()
        for cur_sub_dataset in sub_datasets_list:
            self.has_sub_dataset(cur_sub_dataset)

        sparql_endpoint: Optional[URIRef] = other.get_sparql_endpoint()
        if sparql_endpoint is not None:
            self.has_sparql_endpoint(sparql_endpoint)

        distributions_list: List[Distribution] = other.get_distributions()
        for cur_distribution in distributions_list:
            self.has_distribution(cur_distribution)

    # HAS TITLE
    def get_title(self) -> Optional[str]:
        return self._get_literal(MetadataEntity.iri_title)

    @accepts_only('literal')
    def has_title(self, string: str) -> None:
        """The title of the dataset."""
        self.remove_title()
        self._create_literal(MetadataEntity.iri_title, string)

    def remove_title(self) -> None:
        self.g.remove((self.res, MetadataEntity.iri_title, None))

    # HAS DESCRIPTION
    def get_description(self) -> Optional[str]:
        return self._get_literal(MetadataEntity.iri_description)

    @accepts_only('literal')
    def has_description(self, string: str) -> None:
        """A short textual description of the content of the dataset."""
        self.remove_description()
        self._create_literal(MetadataEntity.iri_description, string)

    def remove_description(self) -> None:
        self.g.remove((self.res, MetadataEntity.iri_description, None))

    # HAS PUBLICATION DATE
    def get_publication_date(self) -> Optional[str]:
        return self._get_literal(MetadataEntity.iri_issued)

    @accepts_only('literal')
    def has_publication_date(self, string: str) -> None:
        """The date of first publication of the dataset."""
        self.remove_publication_date()
        self._create_literal(MetadataEntity.iri_issued, string, XSD.dateTime, False)

    def remove_publication_date(self) -> None:
        self.g.remove((self.res, MetadataEntity.iri_issued, None))

    # HAS MODIFICATION DATE
    def get_modification_date(self) -> Optional[str]:
        return self._get_literal(MetadataEntity.iri_modified)

    @accepts_only('literal')
    def has_modification_date(self, string: str) -> None:
        """The date on which the dataset has been modified."""
        self.remove_modification_date()
        self._create_literal(MetadataEntity.iri_modified, string, XSD.dateTime, False)

    def remove_modification_date(self) -> None:
        self.g.remove((self.res, MetadataEntity.iri_modified, None))

    # HAS KEYWORD
    def get_keywords(self) -> List[str]:
        return self._get_multiple_literals(MetadataEntity.iri_keyword)

    @accepts_only('literal')
    def has_keyword(self, string: str) -> None:
        """A keyword or phrase describing the content of the dataset."""
        self._create_literal(MetadataEntity.iri_keyword, string)

    @accepts_only('literal')
    def remove_keyword(self, string: str = None) -> None:
        if string is not None:
            self.g.remove((self.res, MetadataEntity.iri_keyword, string))
        else:
            self.g.remove((self.res, MetadataEntity.iri_keyword, None))

    # HAS SUBJECT
    def get_subjects(self) -> List[URIRef]:
        return self._get_multiple_uri_references(MetadataEntity.iri_subject)

    @accepts_only('thing')
    def has_subject(self, thing_res: URIRef) -> None:
        """A concept describing the primary subject of the dataset."""
        self.g.add((self.res, MetadataEntity.iri_subject, thing_res))

    @accepts_only('thing')
    def remove_subject(self, thing_res: URIRef = None) -> None:
        if thing_res is not None:
            self.g.remove((self.res, MetadataEntity.iri_subject, thing_res))
        else:
            self.g.remove((self.res, MetadataEntity.iri_subject, None))

    # HAS LANDING PAGE
    def get_landing_page(self) -> Optional[URIRef]:
        return self._get_literal(MetadataEntity.iri_landing_page)

    @accepts_only('thing')
    def has_landing_page(self, thing_res: URIRef) -> None:
        """An HTML page (indicated by its URL) representing a browsable page for the dataset."""
        self.remove_landing_page()
        self.g.add((self.res, MetadataEntity.iri_landing_page, thing_res))

    def remove_landing_page(self) -> None:
        self.g.remove((self.res, MetadataEntity.iri_landing_page, None))

    # HAS SUB-DATASET
    def get_sub_datasets(self) -> List[Dataset]:
        uri_list: List[URIRef] = self._get_multiple_uri_references(MetadataEntity.iri_subset)
        result: List[Dataset] = []
        for uri in uri_list:
            result.append(self.m_set.add_dataset(self.resp_agent, self.source_agent, self.source, uri))
        return result

    @accepts_only('_dataset_')
    def has_sub_dataset(self, obj: Dataset) -> None:
        """A link to a subset of the present dataset."""
        self.g.add((self.res, MetadataEntity.iri_subset, obj.res))

    @accepts_only('_dataset_')
    def remove_sub_dataset(self, dataset_res: Dataset = None) -> None:
        if dataset_res is not None:
            self.g.remove((self.res, MetadataEntity.iri_subset, dataset_res.res))
        else:
            self.g.remove((self.res, MetadataEntity.iri_subset, None))

    # HAS SPARQL ENDPOINT
    def get_sparql_endpoint(self) -> Optional[URIRef]:
        return self._get_uri_reference(MetadataEntity.iri_sparql_endpoint)

    @accepts_only('thing')
    def has_sparql_endpoint(self, thing_res: URIRef) -> None:
        """The link to the SPARQL endpoint for querying the dataset."""
        self.remove_sparql_endpoint()
        self.g.add((self.res, MetadataEntity.iri_sparql_endpoint, thing_res))

    def remove_sparql_endpoint(self) -> None:
        self.g.remove((self.res, MetadataEntity.iri_sparql_endpoint, None))

    # HAS DISTRIBUTION (Distribution)
    def get_distributions(self) -> List[Distribution]:
        uri_list: List[URIRef] = self._get_multiple_uri_references(MetadataEntity.iri_distribution)
        result: List[Distribution] = []
        for uri in uri_list:
            result.append(self.m_set.add_di(self.resp_agent, self.source_agent, self.source, uri))
        return result

    @accepts_only('di')
    def has_distribution(self, obj: Distribution) -> None:
        """A distribution of the dataset."""
        self.g.add((self.res, MetadataEntity.iri_distribution, obj.res))

    @accepts_only('di')
    def remove_distribution(self, di_res: Distribution = None) -> None:
        if di_res is not None:
            self.g.remove((self.res, MetadataEntity.iri_distribution, di_res.res))
        else:
            self.g.remove((self.res, MetadataEntity.iri_distribution, None))



#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright (c) 2016, Silvio Peroni <essepuntato@gmail.com>
#
# Permission to use, copy, modify, and/or distribute this software for any purpose
# with or without fee is hereby granted, provided that the above copyright notice
# and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT,
# OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
# DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
# ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
# SOFTWARE.
from __future__ import annotations

from typing import TYPE_CHECKING

from oc_ocdm.decorators import accepts_only

if TYPE_CHECKING:
    from typing import Optional
    from rdflib import URIRef
    from oc_ocdm.graph.entities.bibliographic.responsible_agent import ResponsibleAgent
from oc_ocdm.graph.graph_entity import GraphEntity
from oc_ocdm.graph.entities.bibliographic_entity import BibliographicEntity


class AgentRole(BibliographicEntity):
    """Agent role (short: ar): a particular role held by an agent with respect to a bibliographic resource."""

    @accepts_only('ar')
    def merge(self, other: AgentRole) -> None:
        super(AgentRole, self).merge(other)

        next_ar: Optional[AgentRole] = other.get_next()
        if next_ar is not None:
            self.has_next(next_ar)

        resp_agent: Optional[ResponsibleAgent] = other.get_is_held_by()
        if resp_agent is not None:
            self.is_held_by(resp_agent)

        role_type: Optional[URIRef] = other.get_role_type()
        if role_type is not None:
            if role_type == GraphEntity.iri_publisher:
                self.create_publisher()
            elif role_type == GraphEntity.iri_author:
                self.create_author()
            elif role_type == GraphEntity.iri_editor:
                self.create_editor()

    # HAS NEXT (AgentRole)
    def get_next(self) -> Optional[AgentRole]:
        uri: Optional[URIRef] = self._get_uri_reference(GraphEntity.iri_has_next)
        if uri is not None:
            return self.g_set.add_ar(self.resp_agent, self.source_agent, self.source, uri)

    @accepts_only('ar')
    def has_next(self, ar_res: AgentRole) -> None:
        """The previous role in a sequence of agent roles of the same type associated with the
        same bibliographic resource (so as to define, for instance, an ordered list of authors).
        """
        self.remove_next()
        self.g.add((self.res, GraphEntity.iri_has_next, ar_res.res))

    def remove_next(self) -> None:
        self.g.remove((self.res, GraphEntity.iri_has_next, None))

    # IS HELD BY (ResponsibleAgent)
    def get_is_held_by(self) -> Optional[ResponsibleAgent]:
        uri: Optional[URIRef] = self._get_uri_reference(GraphEntity.iri_is_held_by)
        if uri is not None:
            return self.g_set.add_ra(self.resp_agent, self.source_agent, self.source, uri)

    @accepts_only('ra')
    def is_held_by(self, ra_res: ResponsibleAgent):
        """The agent holding this role with respect to a particular bibliographic resource.
        """
        self.remove_is_held_by()
        self.g.add((self.res, GraphEntity.iri_is_held_by, ra_res.res))

    def remove_is_held_by(self) -> None:
        self.g.remove((self.res, GraphEntity.iri_is_held_by, None))

    # HAS ROLE TYPE
    def get_role_type(self) -> Optional[URIRef]:
        uri: Optional[URIRef] = self._get_uri_reference(GraphEntity.iri_with_role)
        return uri

    def create_publisher(self) -> None:
        """The specific type of role under consideration (e.g. author, editor or publisher).
        """
        self.remove_role_type()
        self.g.add((self.res, GraphEntity.iri_with_role, GraphEntity.iri_publisher))

    def create_author(self) -> None:
        """The specific type of role under consideration (e.g. author, editor or publisher).
        """
        self.remove_role_type()
        self.g.add((self.res, GraphEntity.iri_with_role, GraphEntity.iri_author))

    def create_editor(self) -> None:
        """The specific type of role under consideration (e.g. author, editor or publisher).
        """
        self.remove_role_type()
        self.g.add((self.res, GraphEntity.iri_with_role, GraphEntity.iri_editor))

    def remove_role_type(self) -> None:
        self.g.remove((self.res, GraphEntity.iri_with_role, None))



#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright (c) 2016, Silvio Peroni <essepuntato@gmail.com>
#
# Permission to use, copy, modify, and/or distribute this software for any purpose
# with or without fee is hereby granted, provided that the above copyright notice
# and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT,
# OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
# DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
# ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
# SOFTWARE.
from __future__ import annotations

from typing import TYPE_CHECKING

from oc_ocdm.decorators import accepts_only

if TYPE_CHECKING:
    from typing import Optional, List
    from rdflib import URIRef
    from oc_ocdm.graph.entities.bibliographic.bibliographic_reference import BibliographicReference
    from oc_ocdm.graph.entities.bibliographic.reference_annotation import ReferenceAnnotation
from oc_ocdm.graph.graph_entity import GraphEntity
from oc_ocdm.graph.entities.bibliographic_entity import BibliographicEntity


class ReferencePointer(BibliographicEntity):
    """Reference pointer (long: in-text reference pointer; short: rp): a textual device (e.g.
       '[1]'), denoting a single bibliographic reference, that is embedded in the text of a
       document within the context of a particular sentence or text chunk. A bibliographic
       reference can be denoted in the text by one or more in-text reference pointers."""

    @accepts_only('rp')
    def merge(self, other: ReferencePointer) -> None:
        super(ReferencePointer, self).merge(other)

        content: Optional[str] = other.get_content()
        if content is not None:
            self.has_content(content)

        next_rp: Optional[ReferencePointer] = other.get_next_rp()
        if next_rp is not None:
            self.has_next_rp(next_rp)

        denoted_be: Optional[BibliographicReference] = other.get_denoted_be()
        if denoted_be is not None:
            self.denotes_be(denoted_be)

        an_list: List[ReferenceAnnotation] = other.get_annotations()
        for cur_an in an_list:
            self.has_annotation(cur_an)

    # HAS REFERENCE POINTER TEXT
    def get_content(self) -> Optional[str]:
        return self._get_literal(GraphEntity.iri_has_content)

    @accepts_only('literal')
    def has_content(self, string: str) -> None:
        """The literal text of the textual device forming an in-text reference pointer and denoting
        a single bibliographic reference (e.g. “[1]”).
        """
        self.remove_content()
        self._create_literal(GraphEntity.iri_has_content, string)

    def remove_content(self) -> None:
        self.g.remove((self.res, GraphEntity.iri_has_content, None))

    # HAS NEXT (ReferencePointer)
    def get_next_rp(self) -> Optional[ReferencePointer]:
        uri: Optional[URIRef] = self._get_uri_reference(GraphEntity.iri_has_next)
        if uri is not None:
            return self.g_set.add_rp(self.resp_agent, self.source_agent, self.source, uri)

    @accepts_only('rp')
    def has_next_rp(self, rp_res: ReferencePointer) -> None:
        """The following in-text reference pointer, when included within a single in-text reference
        pointer list.
        """
        self.remove_next_rp()
        self.g.add((self.res, GraphEntity.iri_has_next, rp_res.res))

    def remove_next_rp(self) -> None:
        self.g.remove((self.res, GraphEntity.iri_has_next, None))

    # DENOTES (BibliographicReference)
    def get_denoted_be(self) -> Optional[BibliographicReference]:
        uri: Optional[URIRef] = self._get_uri_reference(GraphEntity.iri_denotes)
        if uri is not None:
            return self.g_set.add_be(self.resp_agent, self.source_agent, self.source, uri)

    @accepts_only('be')
    def denotes_be(self, be_res: BibliographicReference) -> None:
        """The bibliographic reference included in the list of bibliographic references, denoted by
        the in-text reference pointer.
        """
        self.remove_denoted_be()
        self.g.add((self.res, GraphEntity.iri_denotes, be_res.res))

    def remove_denoted_be(self) -> None:
        self.g.remove((self.res, GraphEntity.iri_denotes, None))

    # HAS ANNOTATION (ReferenceAnnotation)
    def get_annotations(self) -> List[ReferenceAnnotation]:
        uri_list: List[URIRef] = self._get_multiple_uri_references(GraphEntity.iri_has_annotation)
        result: List[ReferenceAnnotation] = []
        for uri in uri_list:
            result.append(self.g_set.add_an(self.resp_agent, self.source_agent, self.source, uri))
        return result

    @accepts_only('an')
    def has_annotation(self, an_res: ReferenceAnnotation) -> None:
        """An annotation characterizing the citation to which the in-text reference pointer relates
        in terms of its citation function (the reason for that citation) specific to the textual
        location of that in-text reference pointer within the citing entity.
        """
        self.g.add((self.res, GraphEntity.iri_has_annotation, an_res.res))

    @accepts_only('an')
    def remove_annotation(self, an_res: ReferenceAnnotation = None) -> None:
        if an_res is not None:
            self.g.remove((self.res, GraphEntity.iri_has_annotation, an_res.res))
        else:
            self.g.remove((self.res, GraphEntity.iri_has_annotation, None))



#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright (c) 2016, Silvio Peroni <essepuntato@gmail.com>
#
# Permission to use, copy, modify, and/or distribute this software for any purpose
# with or without fee is hereby granted, provided that the above copyright notice
# and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT,
# OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
# DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
# ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
# SOFTWARE.
from __future__ import annotations

from typing import TYPE_CHECKING

from oc_ocdm.decorators import accepts_only

if TYPE_CHECKING:
    from typing import List
    from rdflib import URIRef
    from oc_ocdm.graph.entities.identifier import Identifier
from oc_ocdm.graph.graph_entity import GraphEntity

"""
Notes about BibliographicEntity:

    Chill down, everything seems OK here!
"""


class BibliographicEntity(GraphEntity):
    """The base class for each bibliographic entity of the OpenCitations DataModel (OCDM)."""

    def merge(self, other: BibliographicEntity) -> None:
        super(BibliographicEntity, self).merge(other)
        id_list: List[Identifier] = other.get_identifiers()
        for cur_id in id_list:
            self.has_identifier(cur_id)

    # HAS IDENTIFIER
    def get_identifiers(self) -> List[Identifier]:
        uri_list: List[URIRef] = self._get_multiple_uri_references(GraphEntity.iri_has_identifier)
        result: List[Identifier] = []
        for uri in uri_list:
            result.append(self.g_set.add_id(self.resp_agent, self.source_agent, self.source, uri))
        return result

    @accepts_only('id')
    def has_identifier(self, id_res: Identifier) -> None:
        """In addition to the internal dataset identifier assigned to the entity upon initial
        curation (format: [entity short name]/[local identifier]), other external third-party
        identifiers can be specified through this attribute (e.g. DOI, ORCID, PubMedID).
        """
        self.g.add((self.res, GraphEntity.iri_has_identifier, id_res.res))

    @accepts_only('id')
    def remove_identifier(self, id_res: Identifier = None) -> None:
        if id_res is not None:
            self.g.remove((self.res, GraphEntity.iri_has_identifier, id_res.res))
        else:
            self.g.remove((self.res, GraphEntity.iri_has_identifier, None))



#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright (c) 2016, Silvio Peroni <essepuntato@gmail.com>
#
# Permission to use, copy, modify, and/or distribute this software for any purpose
# with or without fee is hereby granted, provided that the above copyright notice
# and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT,
# OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
# DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
# ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
# SOFTWARE.
from __future__ import annotations

from typing import TYPE_CHECKING

from oc_ocdm.decorators import accepts_only

if TYPE_CHECKING:
    from typing import Optional, List
    from rdflib import URIRef
    from oc_ocdm.graph.entities.bibliographic.reference_pointer import ReferencePointer
from oc_ocdm.graph.graph_entity import GraphEntity
from oc_ocdm.graph.entities.bibliographic_entity import BibliographicEntity


class PointerList(BibliographicEntity):
    """Pointer list (short: pl): a textual device (e.g. '[1, 2, 3]' or '[4-9]') which includes a
       number of reference pointers denoting the specific bibliographic references to which
       the list pertains."""

    @accepts_only('pl')
    def merge(self, other: PointerList) -> None:
        super(PointerList, self).merge(other)

        content: Optional[str] = self.get_content()
        if content is not None:
            self.has_content(content)

        rp_list: List[ReferencePointer] = other.get_contained_elements()
        for cur_rp in rp_list:
            self.contains_element(cur_rp)

    # HAS POINTER LIST TEXT
    def get_content(self) -> Optional[str]:
        return self._get_literal(GraphEntity.iri_has_content)

    @accepts_only('literal')
    def has_content(self, string: str) -> None:
        self.remove_content()
        self._create_literal(GraphEntity.iri_has_content, string)

    def remove_content(self) -> None:
        self.g.remove((self.res, GraphEntity.iri_has_content, None))

    # HAS ELEMENT (ReferencePointer)
    def get_contained_elements(self) -> List[ReferencePointer]:
        uri_list: List[URIRef] = self._get_multiple_uri_references(GraphEntity.iri_has_element)
        result: List[ReferencePointer] = []
        for uri in uri_list:
            result.append(self.g_set.add_rp(self.resp_agent, self.source_agent, self.source, uri))
        return result

    @accepts_only('rp')
    def contains_element(self, rp_res: ReferencePointer) -> None:
        """The in-text reference pointer that is part of the in-text reference pointer list present at
        a particular location within the body of the citing work.
        """
        self.g.add((self.res, GraphEntity.iri_has_element, rp_res.res))

    @accepts_only('rp')
    def remove_contained_element(self, rp_res: ReferencePointer = None) -> None:
        if rp_res is not None:
            self.g.remove((self.res, GraphEntity.iri_has_element, rp_res.res))
        else:
            self.g.remove((self.res, GraphEntity.iri_has_element, None))



#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright (c) 2016, Silvio Peroni <essepuntato@gmail.com>
#
# Permission to use, copy, modify, and/or distribute this software for any purpose
# with or without fee is hereby granted, provided that the above copyright notice
# and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT,
# OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
# DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
# ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
# SOFTWARE.
from __future__ import annotations

import os
import re
from datetime import datetime
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Optional, List, Tuple
from urllib.parse import quote

from rdflib import Literal, RDF, URIRef, XSD, Graph


def create_date(date_list: List[Optional[int]] = None) -> Optional[str]:
    string = None
    if date_list is not None:
        l_date_list = len(date_list)
        if l_date_list != 0 and date_list[0] is not None:
            if l_date_list == 3 and \
                    ((date_list[1] is not None and date_list[1] != 1) or
                     (date_list[2] is not None and date_list[2] != 1)):
                string = datetime(date_list[0], date_list[1], date_list[2]).strftime('%Y-%m-%d')
            elif l_date_list == 2 and date_list[1] is not None:
                string = datetime(date_list[0], date_list[1], 1).strftime('%Y-%m')
            else:
                string = datetime(date_list[0], 1, 1,).strftime('%Y')
    return string


def get_datatype_from_iso_8601(string: str) -> Tuple[URIRef, str]:
    # Keep only the "yyyy-mm-dd" part of the string
    string = string[:10]

    try:
        date_parts: List[int] = [int(s) for s in string.split(sep='-', maxsplit=2)]
    except ValueError:
        raise ValueError("The provided date string is not ISO-8601 compliant!")

    num_of_parts: int = len(date_parts)
    if num_of_parts == 3:
        return XSD.date, datetime(*date_parts).strftime('%Y-%m-%d')
    elif num_of_parts == 2:
        return XSD.gYearMonth, datetime(*date_parts, 1).strftime('%Y-%m')
    else:
        return XSD.gYear, datetime(*date_parts, 1, 1).strftime('%Y')


def encode_url(u: str) -> str:
    return quote(u, "://")


def create_literal(g: Graph, res: URIRef, p: URIRef, s: str, dt: URIRef = None, nor: bool = True) -> None:
    string = s
    if not is_string_empty(string):
        g.add((res, p, Literal(string, datatype=dt, normalize=nor)))


def create_type(g: Graph, res: URIRef, res_type: URIRef) -> None:
    g.add((res, RDF.type, res_type))


def is_string_empty(string: str) -> bool:
    return string is None or string.strip() == ""


def get_short_name(res: URIRef) -> str:
    return re.sub("^.+/([a-z][a-z])(/[0-9]+)?$", "\\1", str(res))


def get_prefix(res: URIRef) -> str:
    return re.sub("^.+/[a-z][a-z]/(0[1-9]+0)?([1-9][0-9]*)$", "\\1", str(res))


def get_count(res: URIRef) -> str:
    return re.sub("^.+/[a-z][a-z]/(0[1-9]+0)?([1-9][0-9]*)$", "\\2", str(res))


def get_resource_number(string_iri: str) -> int:
    if "/prov/" in string_iri:
        return int(re.sub(prov_regex, "\\3", string_iri))
    else:
        return int(re.sub(res_regex, "\\3", string_iri))


def find_local_line_id(res: URIRef, n_file_item: int = 1) -> int:
    cur_number = get_resource_number(str(res))

    cur_file_split = 0
    while True:
        if cur_number > cur_file_split:
            cur_file_split += n_file_item
        else:
            cur_file_split -= n_file_item
            break

    return cur_number - cur_file_split


# Variable used in several functions
res_regex: str = "(.+)/(0[1-9]+0)?([1-9][0-9]*)$"
prov_regex: str = "(.+)/(0[1-9]+0)?([1-9][0-9]*)(/prov)/(.+)/([0-9]+)$"


def find_paths(string_iri: str, base_dir: str, base_iri: str, default_dir: str, dir_split: int,
               n_file_item: int, is_json: bool = True) -> Tuple[str, str]:
    """
    This function is responsible for looking for the correct JSON file that contains the data related to the
    resource identified by the variable 'string_iri'. This search takes into account the organisation in
    directories and files, as well as the particular supplier prefix for bibliographic entities, if specified.
    In case no supplier prefix is specified, the 'default_dir' (usually set to "_") is used instead.
    """
    if is_json:
        format_string = ".json"
    else:
        format_string = ".ttl"

    if is_dataset(string_iri):
        cur_dir_path = (base_dir + re.sub("^%s(.*)$" % base_iri, "\\1", string_iri))[:-1]
        # In case of dataset, the file path is different from regular files, e.g.
        # /corpus/br/index.json
        cur_file_path = cur_dir_path + os.sep + "index.json"
        print("is_dataset", cur_dir_path, cur_file_path)

    else:
        cur_number = get_resource_number(string_iri)

        # Find the correct file number where to save the resources
        cur_file_split = 0
        while True:
            if cur_number > cur_file_split:
                cur_file_split += n_file_item
            else:
                break

        # The data have been split in multiple directories and it is not something related
        # with the provenance data of the whole corpus (e.g. provenance agents)
        if dir_split and not string_iri.startswith(base_iri + "prov/"):
            # Find the correct directory number where to save the file
            cur_split = 0
            while True:
                if cur_number > cur_split:
                    cur_split += dir_split
                else:
                    break

            if "/prov/" in string_iri:  # provenance file of a bibliographic entity
                cur_dir_path = base_dir + \
                               re.sub(("^%s" + prov_regex) % base_iri,
                                      ("\\1%s\\2" % os.sep if has_supplier_prefix(string_iri, base_iri) else
                                       "\\1%s%s" % (os.sep, default_dir)), string_iri) + \
                               os.sep + str(cur_split) + os.sep + str(cur_file_split) + os.sep + "prov"
                cur_file_path = cur_dir_path + os.sep + re.sub(
                    ("^%s" + prov_regex) % base_iri, "\\5", string_iri) + format_string
            else:  # regular bibliographic entity
                cur_dir_path = base_dir + \
                               re.sub(("^%s" + res_regex) % base_iri,
                                      ("\\1%s\\2" % os.sep if has_supplier_prefix(string_iri, base_iri) else
                                       "\\1%s%s" % (os.sep, default_dir)),
                                      string_iri) + \
                               os.sep + str(cur_split)

                cur_file_path = cur_dir_path + os.sep + str(cur_file_split) + format_string
        # Enter here if no split is needed
        elif dir_split == 0:
            if "/prov/" in string_iri:
                cur_dir_path = base_dir + \
                               re.sub(("^%s" + prov_regex) % base_iri,
                                      ("\\1%s\\2" % os.sep if has_supplier_prefix(string_iri, base_iri) else
                                       "\\1%s%s" % (os.sep, default_dir)), string_iri) + \
                               os.sep + str(cur_file_split) + os.sep + "prov"
                cur_file_path = cur_dir_path + os.sep + re.sub(
                    ("^%s" + prov_regex) % base_iri, "\\5", string_iri) + format_string
            else:
                cur_dir_path = base_dir + \
                               re.sub(("^%s" + res_regex) % base_iri,
                                      ("\\1%s\\2" % os.sep if has_supplier_prefix(string_iri, base_iri) else
                                       "\\1%s%s" % (os.sep, default_dir)),
                                      string_iri)

                cur_file_path = cur_dir_path + os.sep + str(cur_file_split) + format_string
        # Enter here if the data is about a provenance agent, e.g.,
        # /corpus/prov/
        else:
            cur_dir_path = base_dir + re.sub(("^%s" + res_regex) % base_iri, "\\1", string_iri)
            cur_file_path = cur_dir_path + os.sep + re.sub(res_regex, "\\2\\3", string_iri) + format_string
            print("else:", cur_dir_path, cur_file_path)

    return cur_dir_path, cur_file_path


def has_supplier_prefix(string_iri: str, base_iri: str) -> bool:
    return re.search("^%s[a-z][a-z]/0" % base_iri, string_iri) is not None


def is_dataset(string_iri: str) -> bool:
    return re.search("^.+/[0-9]+(-[0-9]+)?(/[0-9]+)?$", string_iri) is None
